@@ -53,7 +53,7 @@ static int wdt_is_open=0;
  */
  
 static int io=0x240;
-static int irq=14;
+static int irq=11;
 
 #define WD_TIMO (100*60)		/* 1 minute */
 
@@ -155,8 +155,7 @@ void wdt_interrupt(int irq, void *dev_id, struct pt_regs *regs)
 }
 
 
-static long long wdt_llseek(struct inode *inode, struct file *file, long long offset, 
-	int origin)
+static long long wdt_llseek(struct file *file, long long offset, int origin)
 {
 	return -ESPIPE;
 }
@@ -170,8 +169,12 @@ static void wdt_ping(void)
 	outb_p(0, WDT_DC);
 }
 
-static long wdt_write(struct inode *inode, struct file *file, const char *buf, unsigned long count)
+static ssize_t wdt_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
 {
+	/*  Can't seek (pwrite) on this device  */
+	if (ppos != &file->f_pos)
+		return -ESPIPE;
+
 	if(count)
 	{
 		wdt_ping();
@@ -184,13 +187,17 @@ static long wdt_write(struct inode *inode, struct file *file, const char *buf, u
  *	Read reports the temperature in farenheit
  */
  
-static long wdt_read(struct inode *inode, struct file *file, char *buf, unsigned long count)
+static ssize_t wdt_read(struct file *file, char *buf, size_t count, loff_t *ptr)
 {
 	unsigned short c=inb_p(WDT_RT);
 	unsigned char cp;
 	int err;
 	
-	switch(MINOR(inode->i_rdev))
+	/*  Can't seek (pread) on this device  */
+	if (ptr != &file->f_pos)
+		return -ESPIPE;
+
+	switch(MINOR(file->f_dentry->d_inode->i_rdev))
 	{
 		case TEMP_MINOR:
 			err=verify_area(VERIFY_WRITE, buf, 1);
