@@ -157,7 +157,7 @@ __initfunc(int hp_probe1(struct device *dev, int ioaddr))
 				outb_p(irqmap[irq] | HP_RUN, ioaddr + HP_CONFIGURE);
 				outb_p( 0x00 | HP_RUN, ioaddr + HP_CONFIGURE);
 				if (irq == autoirq_report(0)		 /* It's a good IRQ line! */
-					&& request_irq (irq, &ei_interrupt, 0, "hp", NULL) == 0) {
+					&& request_irq (irq, &ei_interrupt, 0, "hp", dev) == 0) {
 					printk(" selecting IRQ %d.\n", irq);
 					dev->irq = *irqp;
 					break;
@@ -171,7 +171,7 @@ __initfunc(int hp_probe1(struct device *dev, int ioaddr))
 	} else {
 		if (dev->irq == 2)
 			dev->irq = 9;
-		if (request_irq(dev->irq, ei_interrupt, 0, "hp", NULL)) {
+		if (request_irq(dev->irq, ei_interrupt, 0, "hp", dev)) {
 			printk (" unable to get IRQ %d.\n", dev->irq);
 			return EBUSY;
 		}
@@ -180,7 +180,7 @@ __initfunc(int hp_probe1(struct device *dev, int ioaddr))
 	/* Allocate dev->priv and fill in 8390 specific dev fields. */
 	if (ethdev_init(dev)) {
 		printk (" unable to get memory for dev->priv.\n");
-		free_irq(dev->irq, NULL);
+		free_irq(dev->irq, dev);
 		return -ENOMEM;
 	}
 
@@ -233,11 +233,10 @@ hp_reset_8390(struct device *dev)
 	outb_p(0x00, hp_base + HP_CONFIGURE);
 	ei_status.txing = 0;
 	/* Pause just a few cycles for the hardware reset to take place. */
-	SLOW_DOWN_IO;
-	SLOW_DOWN_IO;
+	udelay(5);
 
 	outb_p(saved_config, hp_base + HP_CONFIGURE);
-	SLOW_DOWN_IO; SLOW_DOWN_IO;
+	udelay(5);
 
 	if ((inb_p(hp_base+NIC_OFFSET+EN0_ISR) & ENISR_RESET) == 0)
 		printk("%s: hp_reset_8390() did not complete.\n", dev->name);
@@ -435,8 +434,7 @@ cleanup_module(void)
 			int ioaddr = dev->base_addr - NIC_OFFSET;
 			kfree(dev->priv);
 			dev->priv = NULL;
-			free_irq(dev->irq, NULL);
-			irq2dev_map[dev->irq] = NULL;
+			free_irq(dev->irq, dev);
 			release_region(ioaddr, HP_IO_EXTENT);
 			unregister_netdev(dev);
 		}
