@@ -5,12 +5,11 @@
  * other parts of the kernel because they weren't in ksyms.c, others
  * are partial duplicates with added or changed functionality.
  *
- * Note that several functions lock the inode upon which they want
- * to act, most notably those that create directory entries. The
- * unlock operation can take place either by calling fh_unlock within
- * the function directly, or at a later time in fh_put(). So if you
- * notice code paths that apparently fail to unlock the inode, don't
- * worry--they have been taken care of.
+ * Note that several functions dget() the dentry upon which they want
+ * to act, most notably those that create directory entries. Response
+ * dentry's are dput()'d if necessary in the release callback.
+ * So if you notice code paths that apparently fail to dput() the
+ * dentry, don't worry--they have been taken care of.
  *
  * Copyright (C) 1995, 1996, 1997 Olaf Kirch <okir@monad.swb.de>
  */
@@ -26,7 +25,6 @@
 #include <linux/stat.h>
 #include <linux/fcntl.h>
 #include <linux/net.h>
-#include <linux/in.h>
 #include <linux/unistd.h>
 #include <linux/malloc.h>
 #include <linux/in.h>
@@ -38,10 +36,9 @@
 #include <asm/uaccess.h>
 #endif
 
-#define NFSDDBG_FACILITY		NFSDDBG_FILEOP
+extern void fh_update(struct svc_fh*);
 
-/* Symbol not exported */
-static struct super_block *get_super(dev_t dev);
+#define NFSDDBG_FACILITY		NFSDDBG_FILEOP
 
 /* Open mode for nfsd_open */
 #define OPEN_READ	0
@@ -49,7 +46,7 @@ static struct super_block *get_super(dev_t dev);
 
 /* Hack until we have a macro check for mandatory locks. */
 #ifndef IS_ISMNDLK
-#define IS_ISMNDLK(i)	(((i)->i_mode & (S_ISGID|S_ISVTX)) == S_ISGID)
+#define IS_ISMNDLK(i)	(((i)->i_mode & (S_ISGID|S_IXGRP)) == S_ISGID)
 #endif
 
 /* Check for dir entries '.' and '..' */
@@ -63,10 +60,10 @@ static struct super_block *get_super(dev_t dev);
  * add a hash table here.
  */
 struct raparms {
-	struct raparms *	p_next;
+	struct raparms		*p_next;
 	unsigned int		p_count;
-	dev_t			p_dev;
 	ino_t			p_ino;
+	dev_t			p_dev;
 	unsigned long		p_reada,
 				p_ramax,
 				p_raend,
@@ -89,96 +86,76 @@ fs_off_limits(struct super_block *sb)
 }
 
 /*
- * Check whether directory is a mount point
+ * Check whether directory is a mount point, but it is alright if
+ * this is precisely the local mount point being exported.
  */
 static inline int
-nfsd_iscovered(struct inode *inode)
+nfsd_iscovered(struct dentry *dentry, struct svc_export *exp)
 {
-	return inode->i_mount != NULL;
+	return (dentry != dentry->d_covers &&
+		dentry != exp->ex_dentry);
 }
 
 /*
  * Look up one component of a pathname.
+ * N.B. After this call _both_ fhp and resfh need an fh_put
  */
 int
 nfsd_lookup(struct svc_rqst *rqstp, struct svc_fh *fhp, const char *name,
 					int len, struct svc_fh *resfh)
 {
 	struct svc_export	*exp;
-	struct super_block	*sb;
-	struct inode		*dirp, *inode;
-	int			perm, err, dotdot = 0;
+	struct dentry		*dparent;
+	int			err;
 
-	dprintk("nfsd: nfsd_lookup(fh %x/%ld, %s)\n",
-			SVCFH_DEV(fhp), SVCFH_INO(fhp), name);
+	dprintk("nfsd: nfsd_lookup(fh %p, %s)\n", SVCFH_DENTRY(fhp), name);
 
-	/* Obtain inode and export */
-	if ((err = fh_lookup(rqstp, fhp, S_IFDIR, MAY_NOP)) != 0)
-		return err;
-	dirp = fhp->fh_inode;
+	/* Obtain dentry and export. */
+	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_NOP);
+	if (err)
+		goto out;
+
+	dparent = fhp->fh_dentry;
 	exp  = fhp->fh_export;
 
-	/* check permissions before traversing mount-points */
-	perm = nfsd_permission(exp, dirp, MAY_EXEC);
+	/* Fast path... */
+	err = nfsd_permission(exp, dparent, MAY_EXEC);
+	if ((err == 0)					&&
+	    !fs_off_limits(dparent->d_inode->i_sb)	&&
+	    !nfsd_iscovered(dparent, exp)) {
+		struct dentry *dchild;
 
-	dotdot = (len == 2 && name[0] == '.' && name[1] == '.');
-	if (dotdot) {
-		if (dirp == current->fs->root) {
-			dirp->i_count++;
-			*resfh = *fhp;
-			return 0;
-		}
+		/* Lookup the name, but don't follow links */
+		dchild = lookup_dentry(name, dget(dparent), 0);
+		err = PTR_ERR(dchild);
+		if (IS_ERR(dchild))
+			return nfserrno(-err);
 
-		if (dirp->i_dev == exp->ex_dev && dirp->i_ino == exp->ex_ino) {
-			dirp->i_count++;
-			*resfh = *fhp;
-			return 0;
-		}
-	} else if (len == 1 && name[0] == '.') {
-		len = 0;
-	} else if (fs_off_limits(dirp->i_sb)) {
-		/* No lookups on NFS mounts and procfs */
+		/*
+		 * Note: we compose the filehandle now, but as the
+		 * dentry may be negative, it may need to be updated.
+		 */
+		fh_compose(resfh, exp, dchild);
+		return (dchild->d_inode ? 0 : nfserr_noent);
+	}
+
+	/* Slow path... */
+	if (fs_off_limits(dparent->d_inode->i_sb))
 		return nfserr_noent;
-	} else if (nfsd_iscovered(dirp)) {
-		/* broken NFS client */
+	if (nfsd_iscovered(dparent, exp))
 		return nfserr_acces;
-	}
-	if (!dirp->i_op || !dirp->i_op->lookup)
-		return nfserr_notdir;
- 	if (perm != 0)
-		return perm;
-	if (!len) {
-		dirp->i_count++;
-		*resfh = *fhp;
-		return 0;
-	}
-
-	dirp->i_count++;		/* lookup eats the dirp inode */
-	err = dirp->i_op->lookup(dirp, name, len, &inode);
-
-	if (err)
-		return nfserrno(-err);
-
-	/* Note that lookup() has already done a call to iget() so that
-	 * the inode returned never refers to an inode covered by a mount.
-	 * When this has happened, return the covered inode.
-	 */
-	if (!dotdot && (sb = inode->i_sb) && (inode == sb->s_mounted)) {
-		iput(inode);
-		inode = sb->s_covered;
-		inode->i_count++;
-	}
-
-	fh_compose(resfh, exp, inode);
-	return 0;
+out:
+	return err;
 }
 
 /*
  * Set various file attributes.
+ * N.B. After this call fhp needs an fh_put
  */
 int
 nfsd_setattr(struct svc_rqst *rqstp, struct svc_fh *fhp, struct iattr *iap)
 {
+	struct dentry	*dentry;
 	struct inode	*inode;
 	int		accmode = MAY_SATTR;
 	int		ftype = 0;
@@ -191,25 +168,26 @@ nfsd_setattr(struct svc_rqst *rqstp, struct svc_fh *fhp, struct iattr *iap)
 		ftype = S_IFREG;
 
 	/* Get inode */
-	if ((err = fh_lookup(rqstp, fhp, ftype, accmode)) != 0)
-		return err;
+	err = fh_verify(rqstp, fhp, ftype, accmode);
+	if (err)
+		goto out;
 
-	fh_lock(fhp);			/* lock inode */
-	inode = fhp->fh_inode;
+	dentry = fhp->fh_dentry;
+	inode = dentry->d_inode;
 
 	/* The size case is special... */
 	if ((iap->ia_valid & ATTR_SIZE) && S_ISREG(inode->i_mode)) {
 		if (iap->ia_size < inode->i_size) {
-			err = nfsd_permission(fhp->fh_export, inode, MAY_TRUNC);
+			err = nfsd_permission(fhp->fh_export, dentry, MAY_TRUNC);
 			if (err != 0)
-				return err;
+				goto out;
 		}
 		if ((err = get_write_access(inode)) != 0)
 			return nfserrno(-err);
 		inode->i_size = iap->ia_size;
 		if (inode->i_op && inode->i_op->truncate)
 			inode->i_op->truncate(inode);
-		inode->i_dirt = 1;
+		mark_inode_dirty(inode);
 		put_write_access(inode);
 		iap->ia_valid &= ATTR_SIZE;
 		iap->ia_valid |= ATTR_MTIME;
@@ -238,61 +216,75 @@ nfsd_setattr(struct svc_rqst *rqstp, struct svc_fh *fhp, struct iattr *iap)
 	if (iap->ia_valid) {
 		iap->ia_valid |= ATTR_CTIME;
 		iap->ia_ctime = CURRENT_TIME;
-		err = nfsd_notify_change(inode, iap);
+		err = notify_change(inode, iap);
 		if (err)
 			return nfserrno(-err);
 		if (EX_ISSYNC(fhp->fh_export))
-			nfsd_write_inode(inode);
+			write_inode_now(inode);
 	}
-
-	return 0;
+	err = 0;
+out:
+	return err;
 }
 
 /*
  * Open an existing file or directory.
  * The wflag argument indicates write access.
+ * N.B. After this call fhp needs an fh_put
  */
 int
 nfsd_open(struct svc_rqst *rqstp, struct svc_fh *fhp, int type,
 			int wflag, struct file *filp)
 {
+	struct dentry	*dentry;
 	struct inode	*inode;
 	int		access, err;
 
 	access = wflag? MAY_WRITE : MAY_READ;
-	if ((err = fh_lookup(rqstp, fhp, type, access)) != 0)
-		return err;
-	inode = fhp->fh_inode;
+	err = fh_verify(rqstp, fhp, type, access);
+	if (err)
+		goto out;
+
+	dentry = fhp->fh_dentry;
+	inode = dentry->d_inode;
 
 	/* Disallow access to files with the append-only bit set or
-	 * with mandatory locking enabled */
+	 * with mandatory locking enabled
+	 */
+	err = nfserr_perm;
 	if (IS_APPEND(inode) || IS_ISMNDLK(inode))
-		return nfserr_perm;
+		goto out;
 	if (!inode->i_op || !inode->i_op->default_file_ops)
-		return nfserr_perm;
+		goto out;
 
 	if (wflag && (err = get_write_access(inode)) != 0)
-		return nfserrno(-err);
+		goto out_nfserr;
 
 	memset(filp, 0, sizeof(*filp));
 	filp->f_op    = inode->i_op->default_file_ops;
 	filp->f_count = 1;
 	filp->f_flags = wflag? O_WRONLY : O_RDONLY;
 	filp->f_mode  = wflag? FMODE_WRITE : FMODE_READ;
-	filp->f_inode = inode;
+	filp->f_dentry = dentry;
 
-	if (filp->f_op->open) {
+	err = 0;
+	if (filp->f_op && filp->f_op->open) {
 		err = filp->f_op->open(inode, filp);
 		if (err) {
 			if (wflag)
 				put_write_access(inode);
+
+			/* I nearly added put_filp() call here, but this filp
+			 * is really on callers stack frame. -DaveM
+			 */
 			filp->f_count--;
-			return nfserrno(-err);
 		}
 	}
-
-	inode->i_count++;
-	return 0;
+out_nfserr:
+	if (err)
+		err = nfserrno(-err);
+out:
+	return err;
 }
 
 /*
@@ -301,16 +293,18 @@ nfsd_open(struct svc_rqst *rqstp, struct svc_fh *fhp, int type,
 void
 nfsd_close(struct file *filp)
 {
-	struct inode	*inode;
+	struct dentry	*dentry = filp->f_dentry;
+	struct inode	*inode = dentry->d_inode;
 
-	inode = filp->f_inode;
 	if (!inode->i_count)
 		printk(KERN_WARNING "nfsd: inode count == 0!\n");
+	if (!dentry->d_count)
+		printk(KERN_WARNING "nfsd: wheee, %s/%s d_count == 0!\n",
+			dentry->d_parent->d_name.name, dentry->d_name.name);
 	if (filp->f_op && filp->f_op->release)
 		filp->f_op->release(inode, filp);
 	if (filp->f_mode & FMODE_WRITE)
 		put_write_access(inode);
-	iput(inode);
 }
 
 /*
@@ -319,11 +313,12 @@ nfsd_close(struct file *filp)
 void
 nfsd_sync(struct inode *inode, struct file *filp)
 {
-	filp->f_op->fsync(inode, filp);
+	filp->f_op->fsync(filp, filp->f_dentry);
 }
 
 /*
- * Obtain the readahead parameters for the given file
+ * Obtain the readahead parameters for the file
+ * specified by (dev, ino).
  */
 static inline struct raparms *
 nfsd_get_raparms(dev_t dev, ino_t ino)
@@ -331,17 +326,18 @@ nfsd_get_raparms(dev_t dev, ino_t ino)
 	struct raparms	*ra, **rap, **frap = NULL;
 
 	for (rap = &raparm_cache; (ra = *rap); rap = &ra->p_next) {
-		if (ra->p_dev != dev || ra->p_ino != ino) {
-			if (ra->p_count == 0)
-				frap = rap;
-		} else
+		if (ra->p_ino == ino && ra->p_dev == dev)
 			goto found;
+		if (ra->p_count == 0)
+			frap = rap;
 	}
 	if (!frap)
 		return NULL;
 	rap = frap;
 	ra = *frap;
 	memset(ra, 0, sizeof(*ra));
+	ra->p_dev = dev;
+	ra->p_ino = ino;
 found:
 	if (rap != &raparm_cache) {
 		*rap = ra->p_next;
@@ -355,27 +351,27 @@ found:
 /*
  * Read data from a file. count must contain the requested read count
  * on entry. On return, *count contains the number of bytes actually read.
+ * N.B. After this call fhp needs an fh_put
  */
 int
 nfsd_read(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset, char *buf,
 						unsigned long *count)
 {
 	struct raparms	*ra;
-	struct inode	*inode;
-	struct file	file;
-	unsigned long	oldfs;
+	mm_segment_t	oldfs;
 	int		err;
+	struct file	file;
 
-	if ((err = nfsd_open(rqstp, fhp, S_IFREG, OPEN_READ, &file)) != 0)
-		return err;
-	inode = file.f_inode;
-	if (!file.f_op->read) {
-		nfsd_close(&file);
-		return nfserr_perm;
-	}
+	err = nfsd_open(rqstp, fhp, S_IFREG, OPEN_READ, &file);
+	if (err)
+		goto out;
+	err = nfserr_perm;
+	if (!file.f_op->read)
+		goto out_close;
 
 	/* Get readahead parameters */
-	if ((ra = nfsd_get_raparms(inode->i_dev, inode->i_ino)) != NULL) {
+	ra = nfsd_get_raparms(fhp->fh_handle.fh_dev, fhp->fh_handle.fh_ino);
+	if (ra) {
 		file.f_reada = ra->p_reada;
 		file.f_ramax = ra->p_ramax;
 		file.f_raend = ra->p_raend;
@@ -385,7 +381,7 @@ nfsd_read(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset, char *buf,
 	file.f_pos = offset;
 
 	oldfs = get_fs(); set_fs(KERNEL_DS);
-	err = file.f_op->read(file.f_inode, &file, buf, *count);
+	err = file.f_op->read(&file, buf, *count, &file.f_pos);
 	set_fs(oldfs);
 
 	/* Write back readahead params */
@@ -401,17 +397,21 @@ nfsd_read(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset, char *buf,
 		ra->p_count -= 1;
 	}
 
+	if (err >= 0) {
+		*count = err;
+		err = 0;
+	} else 
+		err = nfserrno(-err);
+out_close:
 	nfsd_close(&file);
-
-	if (err < 0)
-		return nfserrno(-err);
-	*count = err;
-	return 0;
+out:
+	return err;
 }
 
 /*
  * Write data to a file.
  * The stable flag requests synchronous writes.
+ * N.B. After this call fhp needs an fh_put
  */
 int
 nfsd_write(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
@@ -419,20 +419,22 @@ nfsd_write(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
 {
 	struct svc_export	*exp;
 	struct file		file;
+	struct dentry		*dentry;
 	struct inode		*inode;
-	unsigned long		oldfs;
-	int			err;
+	mm_segment_t		oldfs;
+	int			err = 0;
 
 	if (!cnt)
-		return 0;
-	if ((err = nfsd_open(rqstp, fhp, S_IFREG, OPEN_WRITE, &file)) != 0)
-		return err;
-	if (!file.f_op->write) {
-		nfsd_close(&file);
-		return nfserr_perm;
-	}
+		goto out;
+	err = nfsd_open(rqstp, fhp, S_IFREG, OPEN_WRITE, &file);
+	if (err)
+		goto out;
+	err = nfserr_perm;
+	if (!file.f_op->write)
+		goto out_close;
 
-	inode = fhp->fh_inode;
+	dentry = file.f_dentry;
+	inode = dentry->d_inode;
 	exp   = fhp->fh_export;
 
 	/*
@@ -450,7 +452,7 @@ nfsd_write(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
 
 	/* Write the data. */
 	oldfs = get_fs(); set_fs(KERNEL_DS);
-	err = file.f_op->write(inode, &file, buf, cnt);
+	err = file.f_op->write(&file, buf, cnt, &file.f_pos);
 	set_fs(oldfs);
 
 	/* clear setuid/setgid flag after write */
@@ -459,7 +461,7 @@ nfsd_write(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
 
 		ia.ia_valid = ATTR_MODE;
 		ia.ia_mode  = inode->i_mode & ~(S_ISUID | S_ISGID);
-		nfsd_notify_change(inode, &ia);
+		notify_change(inode, &ia);
 	}
 
 	fh_unlock(fhp);			/* unlock inode */
@@ -494,169 +496,222 @@ nfsd_write(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
 #endif
 		}
 
-		if (inode->i_dirt) {
+		if (inode->i_state & I_DIRTY) {
 			dprintk("nfsd: write sync %d\n", current->pid);
 			nfsd_sync(inode, &file);
-			nfsd_write_inode(inode);
+			write_inode_now(inode);
 		}
 		wake_up(&inode->i_wait);
 		last_ino = inode->i_ino;
 		last_dev = inode->i_dev;
 	}
 
-	nfsd_close(&file);
-
 	dprintk("nfsd: write complete\n");
-	return (err < 0)? nfserrno(-err) : 0;
+	if (err >= 0)
+		err = 0;
+	else 
+		err = nfserrno(-err);
+out_close:
+	nfsd_close(&file);
+out:
+	return err;
 }
 
 /*
  * Create a file (regular, directory, device, fifo).
  * UNIX sockets not yet implemented.
+ * N.B. Every call to nfsd_create needs an fh_put for _both_ fhp and resfhp
  */
 int
 nfsd_create(struct svc_rqst *rqstp, struct svc_fh *fhp,
 		char *fname, int flen, struct iattr *iap,
 		int type, dev_t rdev, struct svc_fh *resfhp)
 {
-	struct inode	*dirp, *inode = NULL;
+	struct dentry	*dentry, *dchild;
+	struct inode	*dirp;
 	int		err;
 
+	err = nfserr_perm;
 	if (!flen)
-		return nfserr_perm;
-
+		goto out;
 	if (!(iap->ia_valid & ATTR_MODE))
 		iap->ia_mode = 0;
+	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_CREATE);
+	if (err)
+		goto out;
 
-	if ((err = fh_lookup(rqstp, fhp, S_IFDIR, MAY_CREATE)) != 0)
-		return err;
+	dentry = fhp->fh_dentry;
+	dirp = dentry->d_inode;
 
-	fh_lock(fhp);			/* lock directory */
-	dirp = fhp->fh_inode;
-	dirp->i_count++;		/* dirop eats the inode */
+	/* Get all the sanity checks out of the way before we lock the parent. */
+	err = nfserr_notdir;
+	if(!dirp->i_op || !dirp->i_op->lookup)
+		goto out;
+	err = nfserr_perm;
+	if (type == S_IFREG) {
+		if(!dirp->i_op->create)
+			goto out;
+	} else if(type == S_IFDIR) {
+		if(!dirp->i_op->mkdir)
+			goto out;
+	} else if((type == S_IFCHR) || (type == S_IFBLK) || (type == S_IFIFO)) {
+		if(!dirp->i_op->mknod)
+			goto out;
+	} else {
+		goto out;
+	}
 
+	/*
+	 * The response filehandle may have been setup already ...
+	 */
+	if (!resfhp->fh_dverified) {
+		dchild = lookup_dentry(fname, dget(dentry), 0);
+		err = PTR_ERR(dchild);
+		if(IS_ERR(dchild))
+			return nfserrno(-err);
+	} else
+		dchild = resfhp->fh_dentry;
+	/*
+	 * Make sure the child dentry is still negative ...
+	 */
+	if (dchild->d_inode) {
+		printk("nfsd_create: dentry %s/%s not negative!\n",
+			dentry->d_parent->d_name.name, dentry->d_name.name);
+	}
+
+	/* Looks good, lock the directory. */
+	fh_lock(fhp);
 	switch (type) {
 	case S_IFREG:
-		if (!dirp->i_op || !dirp->i_op->create)
-			return nfserr_perm;
-		err = dirp->i_op->create(dirp, fname, flen,
-					 iap->ia_mode, &inode);
+		err = dirp->i_op->create(dirp, dchild, iap->ia_mode);
 		break;
 	case S_IFDIR:
-		if (!dirp->i_op || !dirp->i_op->mkdir)
-			return nfserr_perm;
-		err = dirp->i_op->mkdir(dirp, fname, flen, iap->ia_mode);
+		err = dirp->i_op->mkdir(dirp, dchild, iap->ia_mode);
 		break;
 	case S_IFCHR:
 	case S_IFBLK:
 	case S_IFIFO:
-		if (!dirp->i_op || !dirp->i_op->mknod)
-			return nfserr_perm;
-		err = dirp->i_op->mknod(dirp, fname, flen, iap->ia_mode, rdev);
+		err = dirp->i_op->mknod(dirp, dchild, iap->ia_mode, rdev);
 		break;
-	default:
-		iput(dirp);
-		err = -EACCES;
 	}
-
 	fh_unlock(fhp);
 
 	if (err < 0)
 		return nfserrno(-err);
+	if (EX_ISSYNC(fhp->fh_export))
+		write_inode_now(dirp);
 
 	/*
-	 * If the VFS call doesn't return the inode, look it up now.
+	 * Assemble the file handle for the newly created file,
+	 * or update the filehandle to get the new inode info.
 	 */
-	if (inode == NULL) {
-		dirp->i_count++;
-		err = dirp->i_op->lookup(dirp, fname, flen, &inode);
-		if (err < 0)
-			return -nfserrno(err);	/* Huh?! */
+	if (!resfhp->fh_dverified) {
+		fh_compose(resfhp, fhp->fh_export, dchild);
+	} else {
+		fh_update(resfhp);
 	}
 
-	if (EX_ISSYNC(fhp->fh_export))
-		nfsd_write_inode(dirp);
-
-	/* Assemble the file handle for the newly created file */
-	fh_compose(resfhp, fhp->fh_export, inode);
-
 	/* Set file attributes. Mode has already been set and
 	 * setting uid/gid works only for root. Irix appears to
 	 * send along the gid when it tries to implement setgid
 	 * directories via NFS.
 	 */
-	if ((iap->ia_valid &= (ATTR_UID|ATTR_GID|ATTR_MODE)) != 0) {
-		if ((err = nfsd_setattr(rqstp, resfhp, iap)) != 0) {
-			fh_put(resfhp);
-			return err;
-		}
-	}
-
-	return 0;
+	err = 0;
+	if ((iap->ia_valid &= (ATTR_UID|ATTR_GID|ATTR_MODE)) != 0)
+		err = nfsd_setattr(rqstp, resfhp, iap);
+out:
+	return err;
 }
 
 /*
  * Truncate a file.
  * The calling routines must make sure to update the ctime
  * field and call notify_change.
+ *
+ * XXX Nobody calls this thing? -DaveM
+ * N.B. After this call fhp needs an fh_put
  */
 int
 nfsd_truncate(struct svc_rqst *rqstp, struct svc_fh *fhp, unsigned long size)
 {
+	struct dentry	*dentry;
 	struct inode	*inode;
+	struct iattr	newattrs;
 	int		err;
 
-	if ((err = fh_lookup(rqstp, fhp, S_IFREG, MAY_WRITE|MAY_TRUNC)) != 0)
-		return err;
+	err = fh_verify(rqstp, fhp, S_IFREG, MAY_WRITE | MAY_TRUNC);
+	if (err)
+		goto out;
 
-	fh_lock(fhp);			/* lock inode if not yet locked */
-	inode = fhp->fh_inode;
+	dentry = fhp->fh_dentry;
+	inode = dentry->d_inode;
 
-	if ((err = get_write_access(inode)) != 0)
-		return nfserrno(-err);
-	inode->i_size = size;
-	if (inode->i_op && inode->i_op->truncate)
-		inode->i_op->truncate(inode);
-	inode->i_dirt = 1;
+	err = get_write_access(inode);
+	if (err)
+		goto out_nfserr;
+
+	/* Things look sane, lock and do it. */
+	fh_lock(fhp);
+	newattrs.ia_size = size;
+	newattrs.ia_valid = ATTR_SIZE | ATTR_CTIME;
+	err = notify_change(inode, &newattrs);
+	if (!err) {
+		vmtruncate(inode, size);
+		if (inode->i_op && inode->i_op->truncate)
+			inode->i_op->truncate(inode);
+	}
 	put_write_access(inode);
-
 	fh_unlock(fhp);
-
-	return 0;
+out_nfserr:
+	if (err)
+		err = nfserrno(-err);
+out:
+	return err;
 }
 
 /*
  * Read a symlink. On entry, *lenp must contain the maximum path length that
  * fits into the buffer. On return, it contains the true length.
+ * N.B. After this call fhp needs an fh_put
  */
 int
 nfsd_readlink(struct svc_rqst *rqstp, struct svc_fh *fhp, char *buf, int *lenp)
 {
+	struct dentry	*dentry;
 	struct inode	*inode;
-	unsigned long	oldfs;
+	mm_segment_t	oldfs;
 	int		err;
 
-	if ((err = fh_lookup(rqstp, fhp, S_IFLNK, MAY_READ)) != 0)
-		return err;
-	inode = fhp->fh_inode;
+	err = fh_verify(rqstp, fhp, S_IFLNK, MAY_READ);
+	if (err)
+		goto out;
 
+	dentry = fhp->fh_dentry;
+	inode = dentry->d_inode;
+
+	err = nfserr_inval;
 	if (!inode->i_op || !inode->i_op->readlink)
-		return nfserr_io;
+		goto out;
 
-	inode->i_count++;
+	UPDATE_ATIME(inode);
 	oldfs = get_fs(); set_fs(KERNEL_DS);
 	err = inode->i_op->readlink(inode, buf, *lenp);
 	set_fs(oldfs);
 
 	if (err < 0)
-		return nfserrno(-err);
-	*lenp = err;
+		err = nfserrno(-err);
+	else {
+		*lenp = err;
+		err = 0;
+	}
 
-	return 0;
+out:
+	return err;
 }
 
 /*
  * Create a symlink and look up its inode
+ * N.B. After this call _both_ fhp and resfhp need an fh_put
  */
 int
 nfsd_symlink(struct svc_rqst *rqstp, struct svc_fh *fhp,
@@ -664,161 +719,260 @@ nfsd_symlink(struct svc_rqst *rqstp, struct svc_fh *fhp,
 				char *path,  int plen,
 				struct svc_fh *resfhp)
 {
-	struct inode	*dirp, *inode;
+	struct dentry	*dentry, *dnew;
+	struct inode	*dirp;
 	int		err;
 
+	err = nfserr_noent;
 	if (!flen || !plen)
-		return nfserr_noent;
-
-	if ((err = fh_lookup(rqstp, fhp, S_IFDIR, MAY_CREATE)) != 0)
-		return err;
-
-	dirp = fhp->fh_inode;
-	if (nfsd_iscovered(dirp))
-		return nfserr_perm;
-	if (!dirp->i_op || !dirp->i_op->symlink)
-		return nfserr_perm;
-
-	fh_lock(fhp);			/* lock inode */
-	dirp->i_count++;
-	err = dirp->i_op->symlink(dirp, fname, flen, path);
-	fh_unlock(fhp);			/* unlock inode */
+		goto out;
 
+	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_CREATE);
 	if (err)
-		return nfserrno(-err);
+		goto out;
+	dentry = fhp->fh_dentry;
 
-	if (EX_ISSYNC(fhp->fh_export))
-		nfsd_write_inode(dirp);
+	err = nfserr_perm;
+	if (nfsd_iscovered(dentry, fhp->fh_export))
+		goto out;
+	dirp = dentry->d_inode;
+	if (!dirp->i_op	|| !dirp->i_op->symlink)
+		goto out;
 
-	/*
-	 * Okay, now look up the inode of the new symlink.
-	 */
-	dirp->i_count++;		/* lookup eats the dirp inode */
-	err = dirp->i_op->lookup(dirp, fname, flen, &inode);
+	dnew = lookup_dentry(fname, dget(dentry), 0);
+	err = PTR_ERR(dnew);
+	if (IS_ERR(dnew))
+		goto out_nfserr;
+
+	err = -EEXIST;
+	if (!dnew->d_inode) {
+		fh_lock(fhp);
+		err = dirp->i_op->symlink(dirp, dnew, path);
+		fh_unlock(fhp);
+		if (!err) {
+			if (EX_ISSYNC(fhp->fh_export))
+				write_inode_now(dirp);
+		}
+	}
+	fh_compose(resfhp, fhp->fh_export, dnew);
+
+out_nfserr:
 	if (err)
-		return nfserrno(-err);
-
-	fh_compose(resfhp, fhp->fh_export, inode);
-	return 0;
+		err = nfserrno(-err);
+out:
+	return err;
 }
 
 /*
  * Create a hardlink
+ * N.B. After this call _both_ ffhp and tfhp need an fh_put
  */
 int
 nfsd_link(struct svc_rqst *rqstp, struct svc_fh *ffhp,
 				char *fname, int len, struct svc_fh *tfhp)
 {
+	struct dentry	*ddir, *dnew;
 	struct inode	*dirp, *dest;
 	int		err;
 
-	if ((err = fh_lookup(rqstp, ffhp, S_IFDIR, MAY_CREATE) != 0) ||
-	    (err = fh_lookup(rqstp, tfhp, S_IFREG, MAY_NOP)) != 0)
-		return err;
-	dirp = ffhp->fh_inode;
-	dest = tfhp->fh_inode;
+	err = fh_verify(rqstp, ffhp, S_IFDIR, MAY_CREATE);
+	if (err)
+		goto out;
+	err = fh_verify(rqstp, tfhp, S_IFREG, MAY_NOP);
+	if (err)
+		goto out;
 
+	ddir = ffhp->fh_dentry;
+	dirp = ddir->d_inode;
+
+	dnew = lookup_dentry(fname, dget(ddir), 0);
+	err = PTR_ERR(dnew);
+	if (IS_ERR(dnew))
+		goto out_nfserr;
+
+	err = -EEXIST;
+	if (dnew->d_inode)
+		goto dput_and_out;
+	dest = tfhp->fh_dentry->d_inode;
+
+	err = -EPERM;
 	if (!len)
-		return nfserr_perm;
-	if (nfsd_iscovered(dirp))
-		return nfserr_acces;
-	if (dirp->i_dev != dest->i_dev)
-		return nfserr_acces;	/* FIXME: nxdev for NFSv3 */
-	if (IS_IMMUTABLE(dest) /* || IS_APPEND(dest) */ )
-		return nfserr_perm;
-	if (!dirp->i_op || !dirp->i_op->link)
-		return nfserr_perm;
+		goto dput_and_out;
 
-	fh_lock(ffhp);			/* lock directory inode */
-	dirp->i_count++;
-	err = dirp->i_op->link(dest, dirp, fname, len);
-	fh_unlock(ffhp);		/* unlock inode */
+	err = -EACCES;
+	if (nfsd_iscovered(ddir, ffhp->fh_export))
+		goto dput_and_out;
+	if (dirp->i_dev != dest->i_dev)
+		goto dput_and_out;	/* FIXME: nxdev for NFSv3 */
+
+	err = -EPERM;
+	if (IS_IMMUTABLE(dest) /* || IS_APPEND(dest) */ )
+		goto dput_and_out;
+	if (!dirp->i_op || !dirp->i_op->link)
+		goto dput_and_out;
+
+	fh_lock(ffhp);
+	err = dirp->i_op->link(dest, dirp, dnew);
+	fh_unlock(ffhp);
 
 	if (!err && EX_ISSYNC(ffhp->fh_export)) {
-		nfsd_write_inode(dirp);
-		nfsd_write_inode(dest);
+		write_inode_now(dirp);
+		write_inode_now(dest);
 	}
+dput_and_out:
+	dput(dnew);
+out_nfserr:
+	if (err)
+		err = nfserrno(-err);
+out:
+	return err;
+}
 
-	return err? nfserrno(-err) : 0;
+/*
+ * This follows the model of double_lock() in the VFS.
+ */
+static inline void nfsd_double_down(struct semaphore *s1, struct semaphore *s2)
+{
+	if (s1 != s2) {
+		if ((unsigned long) s1 < (unsigned long) s2) {
+			struct semaphore *tmp = s1;
+			s1 = s2;
+			s2 = tmp;
+		}
+		down(s1);
+	}
+	down(s2);
+}
+
+static inline void nfsd_double_up(struct semaphore *s1, struct semaphore *s2)
+{
+	up(s1);
+	if (s1 != s2)
+		up(s2);
 }
 
 /*
  * Rename a file
+ * N.B. After this call _both_ ffhp and tfhp need an fh_put
  */
 int
 nfsd_rename(struct svc_rqst *rqstp, struct svc_fh *ffhp, char *fname, int flen,
 			    struct svc_fh *tfhp, char *tname, int tlen)
 {
+	struct dentry	*fdentry, *tdentry, *odentry, *ndentry;
 	struct inode	*fdir, *tdir;
 	int		err;
 
-	if ((err = fh_lookup(rqstp, ffhp, S_IFDIR, MAY_REMOVE) != 0)
-	 || (err = fh_lookup(rqstp, tfhp, S_IFDIR, MAY_CREATE)) != 0)
-		return err;
-	fdir = ffhp->fh_inode;
-	tdir = tfhp->fh_inode;
+	err = fh_verify(rqstp, ffhp, S_IFDIR, MAY_REMOVE);
+	if (err)
+		goto out;
+	err = fh_verify(rqstp, tfhp, S_IFDIR, MAY_CREATE);
+	if (err)
+		goto out;
 
+	fdentry = ffhp->fh_dentry;
+	fdir = fdentry->d_inode;
+
+	tdentry = tfhp->fh_dentry;
+	tdir = tdentry->d_inode;
+
+	/* N.B. We shouldn't need this ... dentry layer handles it */
+	err = nfserr_perm;
 	if (!flen || (fname[0] == '.' && 
 	    (flen == 1 || (flen == 2 && fname[1] == '.'))) ||
 	    !tlen || (tname[0] == '.' && 
 	    (tlen == 1 || (tlen == 2 && tname[1] == '.'))))
-		return nfserr_perm;
+		goto out;
 
+	err = -EXDEV;
 	if (fdir->i_dev != tdir->i_dev)
-		return nfserr_acces;	/* nfserr_nxdev */
+		goto out_nfserr;
+	err = -EPERM;
 	if (!fdir->i_op || !fdir->i_op->rename)
-		return nfserr_perm;
+		goto out_nfserr;
 
-	fh_lock(tfhp);			/* lock destination directory */
-	tdir->i_count++;
-	fdir->i_count++;
-	err = fdir->i_op->rename(fdir, fname, flen, tdir, tname, tlen, 0);
-	fh_unlock(tfhp);		/* unlock inode */
+	odentry = lookup_dentry(fname, dget(fdentry), 0);
+	err = PTR_ERR(odentry);
+	if (IS_ERR(odentry))
+		goto out_nfserr;
 
+	ndentry = lookup_dentry(tname, dget(tdentry), 0);
+	err = PTR_ERR(ndentry);
+	if (IS_ERR(ndentry))
+		goto out_dput_old;
+
+	/*
+	 * Lock the parent directories.
+	 */
+	nfsd_double_down(&tdir->i_sem, &fdir->i_sem);
+	/* N.B. check for parent changes after locking?? */
+
+	err = fdir->i_op->rename(fdir, odentry, tdir, ndentry);
 	if (!err && EX_ISSYNC(tfhp->fh_export)) {
-		nfsd_write_inode(fdir);
-		nfsd_write_inode(tdir);
+		write_inode_now(fdir);
+		write_inode_now(tdir);
 	}
+	nfsd_double_up(&tdir->i_sem, &fdir->i_sem);
+	dput(ndentry);
 
-	return err? nfserrno(-err) : 0;
+out_dput_old:
+	dput(odentry);
+out_nfserr:
+	if (err)
+		err = nfserrno(-err);
+out:
+	return err;
 }
 
 /*
  * Unlink a file or directory
+ * N.B. After this call fhp needs an fh_put
  */
 int
 nfsd_unlink(struct svc_rqst *rqstp, struct svc_fh *fhp, int type,
 				char *fname, int flen)
 {
+	struct dentry	*dentry, *rdentry;
 	struct inode	*dirp;
 	int		err;
 
+	/* N.B. We shouldn't need this test ... handled by dentry layer */
+	err = nfserr_acces;
 	if (!flen || isdotent(fname, flen))
-		return nfserr_acces;
+		goto out;
+	err = fh_verify(rqstp, fhp, S_IFDIR, MAY_REMOVE);
+	if (err)
+		goto out;
 
-	if ((err = fh_lookup(rqstp, fhp, S_IFDIR, MAY_REMOVE)) != 0)
-		return err;
+	dentry = fhp->fh_dentry;
+	dirp = dentry->d_inode;
 
-	fh_lock(fhp);			/* lock inode */
-	dirp = fhp->fh_inode;
+	rdentry = lookup_dentry(fname, dget(dentry), 0);
+	err = PTR_ERR(rdentry);
+	if (IS_ERR(rdentry))
+		goto out_nfserr;
 
+	fh_lock(fhp);
 	if (type == S_IFDIR) {
-		if (!dirp->i_op || !dirp->i_op->rmdir)
-			return nfserr_notdir;
-		dirp->i_count++;
-		err = dirp->i_op->rmdir(dirp, fname, flen);
-	} else {	/* other than S_IFDIR */
-		if (!dirp->i_op || !dirp->i_op->unlink)
-			return nfserr_perm;
-		dirp->i_count++;
-		err = dirp->i_op->unlink(dirp, fname, flen);
+		err = -ENOTDIR;
+		if (dirp->i_op && dirp->i_op->rmdir)
+			err = dirp->i_op->rmdir(dirp, rdentry);
+	} else {
+		err = -EPERM;
+		if (dirp->i_op && dirp->i_op->unlink)
+			err = dirp->i_op->unlink(dirp, rdentry);
 	}
+	fh_unlock(fhp);
 
-	fh_unlock(fhp);			/* unlock inode */
+	dput(rdentry);
 	if (!err && EX_ISSYNC(fhp->fh_export))
-		nfsd_write_inode(dirp);
+		write_inode_now(dirp);
 
-	return err? nfserrno(-err) : 0;
+out_nfserr:
+	if (err)
+		err = nfserrno(-err);
+out:
+	return err;
 }
 
 /*
@@ -829,6 +983,7 @@ nfsd_readdir(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
 			encode_dent_fn func, u32 *buffer, int *countp)
 {
 	struct readdir_cd cd;
+	struct inode	*inode;
 	struct file	file;
 	u32		*p;
 	int		oldlen, eof, err;
@@ -856,6 +1011,7 @@ nfsd_readdir(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
 	 * readdir() is not guaranteed to fill up the entire buffer, but
 	 * may choose to do less.
 	 */
+	inode = file.f_dentry->d_inode;
 	do {
 		oldlen = cd.buflen;
 
@@ -864,7 +1020,7 @@ nfsd_readdir(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
 			file.f_inode->i_dev, file.f_inode->i_ino,
 			(int) file.f_pos, (int) oldlen, (int) cd.buflen);
 		 */
-		err = file.f_op->readdir(file.f_inode, &file,
+		err = file.f_op->readdir(&file,
 					 &cd, (filldir_t) func);
 
 		if (err < 0) {
@@ -900,36 +1056,44 @@ nfsd_readdir(struct svc_rqst *rqstp, struct svc_fh *fhp, loff_t offset,
 
 /*
  * Get file system stats
+ * N.B. After this call fhp needs an fh_put
  */
 int
 nfsd_statfs(struct svc_rqst *rqstp, struct svc_fh *fhp, struct statfs *stat)
 {
+	struct dentry		*dentry;
 	struct inode		*inode;
 	struct super_block	*sb;
-	unsigned long		oldfs;
+	mm_segment_t		oldfs;
 	int			err;
 
-	if ((err = fh_lookup(rqstp, fhp, 0, MAY_NOP)) != 0)
-		return err;
-	inode = fhp->fh_inode;
+	err = fh_verify(rqstp, fhp, 0, MAY_NOP);
+	if (err)
+		goto out;
+	dentry = fhp->fh_dentry;
+	inode = dentry->d_inode;
 
+	err = nfserr_io;
 	if (!(sb = inode->i_sb) || !sb->s_op->statfs)
-		return nfserr_io;
+		goto out;
 
 	oldfs = get_fs();
 	set_fs (KERNEL_DS);
 	sb->s_op->statfs(sb, stat, sizeof(*stat));
 	set_fs (oldfs);
+	err = 0;
 
-	return 0;
+out:
+	return err;
 }
 
 /*
  * Check for a user's access permissions to this inode.
  */
 int
-nfsd_permission(struct svc_export *exp, struct inode *inode, int acc)
+nfsd_permission(struct svc_export *exp, struct dentry *dentry, int acc)
 {
+	struct inode	*inode = dentry->d_inode;
 	int		err;
 
 	if (acc == MAY_NOP)
@@ -954,7 +1118,7 @@ nfsd_permission(struct svc_export *exp, struct inode *inode, int acc)
 	if (acc & (MAY_WRITE | MAY_SATTR | MAY_TRUNC)) {
 		if (EX_RDONLY(exp) || IS_RDONLY(inode))
 			return nfserr_rofs;
-		if (S_ISDIR(inode->i_mode) && nfsd_iscovered(inode))
+		if (S_ISDIR(inode->i_mode) && nfsd_iscovered(dentry, exp))
 			return nfserr_perm;
 		if (/* (acc & MAY_WRITE) && */ IS_IMMUTABLE(inode))
 			return nfserr_perm;
@@ -983,100 +1147,6 @@ nfsd_permission(struct svc_export *exp, struct inode *inode, int acc)
 	return err? nfserrno(-err) : 0;
 }
 
-/*
- * Look up the inode for a given FH.
- */
-struct inode *
-nfsd_iget(dev_t dev, ino_t ino)
-{
-	struct super_block	*sb;
-
-	if (!(sb = get_super(dev)) || fs_off_limits(sb))
-		return NULL;
-	return __iget(sb, ino, 0);
-}
-
-/*
- * Write the inode if dirty (copy of fs/inode.c:write_inode)
- */
-void
-nfsd_write_inode(struct inode *inode)
-{
-	struct super_block	*sb = inode->i_sb;
-
-	if (!inode->i_dirt)
-		return;
-	while (inode->i_lock) {
-		sleep_on(&inode->i_wait);
-		if (!inode->i_dirt)
-			return;
-	}
-	if (!sb || !sb->s_op || !sb->s_op->write_inode) {
-		inode->i_dirt = 0;
-		return;
-	}
-	inode->i_lock = 1;
-	sb->s_op->write_inode(inode);
-	inode->i_lock = 0;
-	wake_up(&inode->i_wait);
-}
-
-/*
- * Look up the root inode of the parent fs.
- * We have to go through iget in order to allow for wait_on_inode.
- */
-int
-nfsd_parentdev(dev_t* devp)
-{
-	struct super_block	*sb;
-
-	if (!(sb = get_super(*devp)) || !sb->s_covered)
-		return 0;
-	if (*devp == sb->s_covered->i_dev)
-		return 0;
-	*devp = sb->s_covered->i_dev;
-	return 1;
-}
-
-/* Duplicated here from fs/super.c because it's not exported */
-static struct super_block *
-get_super(dev_t dev)
-{
-	struct super_block *s;
-
-	if (!dev)
-		return NULL;
-	s = 0 + super_blocks;
-	while (s < NR_SUPER + super_blocks)
-		if (s->s_dev == dev) {
-			wait_on_super(s);
-			if (s->s_dev == dev)
-				return s;
-			s = 0 + super_blocks;
-		} else
-			s++;
-	return NULL;
-}
-
-/*
- * This is a copy from fs/inode.c because it wasn't exported.
- */
-int
-nfsd_notify_change(struct inode *inode, struct iattr *attr)
-{
-	int retval;
-
-	if (inode->i_sb && inode->i_sb->s_op &&
-	    inode->i_sb->s_op->notify_change)
-		return inode->i_sb->s_op->notify_change(inode, attr);
-
-	if ((retval = inode_change_ok(inode, attr)) != 0)
-		return retval;
-
-	inode_setattr(inode, attr);
-	return 0;
-}
-
 /*
  * Initialize readahead param cache
  */
